// @flow
/*::
type Severity = 0 | 1 | 2;
type RuleLevel = Severity | 'off' | 'warn' | 'error';
type RuleLevelAndOptions = [RuleLevel, any];
type ParserOptions = {
  +ecmaVersion?:
    | 3
    | 5
    | 6
    | 7
    | 8
    | 9
    | 10
    | 11
    | 2015
    | 2016
    | 2017
    | 2018
    | 2019
    | 2020,
  +sourceType?: 'script' | 'module',
  +ecmaFeatures?: {
    +globalReturn?: boolean,
    +impliedStrict?: boolean,
    +jsx?: boolean,
    +experimentalObjectRestSpread?: boolean,
    +[key: string]: any,
    ...
  },
  +[key: string]: any,
  ...
};

type RuleOverride = {
  +extends?: string | string[],
  +excludedFiles?: string[],
  +files?: string[],
  +rules?: { +[name: string]: RuleLevel | RuleLevelAndOptions },
  ...
};

type Preset = {
  +env?: { +[name: string]: boolean, ... },
  +extends?: string | string[],
  +globals?: { +[name: string]: boolean, ... },
  +overrides?: RuleOverride[],
  +parser?: string,
  +parserOptions?: ParserOptions,
  +plugins?: string[],
  +processor?: string,
  +rules?: { +[name: string]: RuleLevel | RuleLevelAndOptions, ... },
  +settings?: { +[name: string]: any },
  ...
};
*/

const extendForTests = require('./overrides-tests');
const extendForTypescript = require('./overrides-typescript');

const mergePresets = (...presets /*: Array<Preset>*/) /* : Preset */ => {
  const flatPresets = presets.filter(Boolean).map(({ extends: ancestors, ...preset }) => {
    if (ancestors) {
      const ancestorConfigs = (Array.isArray(ancestors) ? ancestors : [ancestors]).filter(
        Boolean,
      );
      return mergePresets(...ancestors, preset);
    } else {
      return preset;
    }
  });

  // last one if exists
  const parser = flatPresets
    .filter(preset => preset.parser)
    .map(preset => preset.parser)
    .slice(-1)[0];

  // flatten
  const overrides = flatPresets.flatMap(preset => preset.overrides || []);

  // merge keys
  const env = flatPresets
    .map(preset => preset.env || {})
    .reduce((res, x) => ({ ...res, ...x }), {});
  const parserOptions = flatPresets
    .map(preset => preset.parserOptions || {})
    .reduce((res, x) => ({ ...res, ...x }), {});
  const rules = flatPresets
    .map(preset => preset.rules || {})
    .reduce((res, x) => ({ ...res, ...x }), {});
  const settings = flatPresets
    .map(preset => preset.settings || {})
    .reduce((res, x) => ({ ...res, ...x }), {});
  const globals = flatPresets
    .map(preset => preset.global || {})
    .reduce((res, x) => ({ ...res, ...x }), {});

  return { env, globals, parser, parserOptions, settings, rules, overrides };
};

const presets = [
  require('eslint-config-airbnb'),
  require('eslint-config-airbnb/hooks'),
  require('eslint-plugin-import/config/errors'),
  require('eslint-plugin-import/config/warnings'),
  require('eslint-plugin-react').configs.recommended,
  require('eslint-plugin-sonarjs').configs.recommended,
  require('eslint-plugin-unicorn').configs.recommended,
  require('eslint-config-prettier'),
  require('eslint-config-prettier/babel'),
  require('eslint-config-prettier/react'),
  require('eslint-config-prettier/unicorn'),
];

const plugins = [
  'eslint-plugin-babel',
  'eslint-plugin-import',
  'eslint-plugin-sonarjs',
  'eslint-plugin-unicorn',
  'eslint-plugin-jsx-a11y',
  'eslint-plugin-react',
  'eslint-plugin-react-hooks',
  'eslint-plugin-flowtype',
  'eslint-plugin-prettier',
];
const finalConfig = {
  extends: [
    require('eslint-plugin-flowtype').configs.recommended,
    require('eslint-plugin-prettier').configs.recommended,
  ],
  plugins: [...plugins, require('eslint-plugin-flowtype')],
  settings: {
    'import/extensions': ['.js', '.mjs', '.ts', '.jsx', '.tsx'],
  },

  rules: {
    'no-else-return': 'off',
    'no-console': 'warn',
    'no-debugger': 'warn',
    'no-multiple-empty-lines': ['error', { max: 3, maxEOF: 0 }],
    'no-underscore-dangle': 'off',
    'no-unused-expressions': [
      'error',
      { allowShortCircuit: true, allowTernary: true, allowTaggedTemplates: true },
    ],
    'no-useless-computed-key': 'error',
    'no-useless-constructor': 'error',
    'no-useless-rename': [
      'warn',
      { ignoreDestructuring: false, ignoreImport: false, ignoreExport: false },
    ],
    'padded-blocks': ['error', { blocks: 'never', switches: 'never', classes: 'never' }],
    'prefer-const': 'error',
    camelcase: 'off',
    'babel/camelcase': 'warn',
    'import/default': 'warn',
    'import/named': 'warn',
    'import/namespace': 'error',
    'import/extensions': ['warn', 'never'],
    'sonarjs/no-duplicate-string': ['error', 5],
    'unicorn/filename-case': 'off',
    'unicorn/catch-error-name': ['warn', { name: 'err' }],
    'unicorn/prevent-abbreviations': 'off',
    'flowtype/define-flow-type': 'error',
    'flowtype/require-valid-file-annotation': 'error',
    'flowtype/use-flow-type': 'error',
    'jsx-a11y/anchor-is-valid': 'off', // next.js
    'jsx-a11y/aria-props': 'error',
    'jsx-a11y/aria-role': 'error',
    'jsx-a11y/heading-has-content': 'error',
    'jsx-a11y/href-no-hash': 'off',
    'jsx-a11y/label-has-for': 'off',
    'jsx-a11y/mouse-events-have-key-events': 'error',
    'jsx-a11y/role-has-required-aria-props': 'error',
    'jsx-a11y/role-supports-aria-props': 'error',
    'jsx-a11y/img-redundant-alt': 'error',
    'jsx-a11y/no-access-key': 'error',
    'react/jsx-indent': 'off',
    'react/require-default-props': 'off',
    'react/forbid-prop-types': ['error', { forbid: ['any'] }],
    'react/default-props-match-prop-types': 'off', // flow/ts
    'react/jsx-curly-brace-presence': 'off', // styled-jsx
    'react/jsx-filename-extension': ['error', { extensions: ['.js', '.mjs', '.tsx'] }],
    'react/no-array-index-key': 'off',
    'react/no-unused-prop-types': 'off', // flow/ts
    'react/jsx-indent-props': 'off',
    'react/jsx-wrap-multilinis': 'off',
    'react/sort-comp': 'off',
    'react/jsx-props-no-spreading': 'off', // todo
    'react-hooks/rules-of-hooks': 'error',
    'react-hooks/exhaustive-deps': 'warn',
    // 'emotion/syntax-preference': ['error', 'object'],
    // 'graphql/template-strings': ['error', { env: 'apollo', validators: 'all' }],
  },
};

const config = mergePresets(...presets, finalConfig);
config.plugins = plugins;

module.exports = extendForTests(extendForTypescript(config));
